.drawing .text-button {
  font-family: "FontAwesome 5 Free Solid";
  padding: 4px;
}

.drawing .text-button radio,
.color-box .text-button radio {
  padding: 0;
}

.color-box button {
  padding: 6px 10px;
}

.color-box image {
  border-radius: 50px;
}

.color-box .color-red image {
  background-color: rgb(255, 0, 0);
}

.color-box .color-green image {
  background-color: rgb(0, 255, 0);
}

.color-box .color-blue image {
  background-color: rgb(0, 0, 255);
}
